use crate::vm::object::DvmObject;
use crate::vm::value::DvmValue;
use crate::vm::DvmContext;
use std::cell::RefCell;
use std::rc::Rc;

/// ```deimos
/// // Signature
/// print(m: &(Byte | Int | Long | Double | USize | Boolean | Object | Empty))
/// ```
pub fn dm_print(context: &mut DvmContext) {
    print!("{}", get_string(context.pop_stack()));
}

pub fn dm_println(context: &mut DvmContext) {
    println!("{}", get_string(context.pop_stack()));
}

fn get_string(dvm_value: DvmValue) -> String {
    match dvm_value {
        DvmValue::Byte(b) => b.to_string(),
        DvmValue::Int(i) => i.to_string(),
        DvmValue::Long(l) => l.to_string(),
        DvmValue::Double(d) => d.to_string(),
        DvmValue::Boolean(b) => b.to_string(),
        DvmValue::Object(object) => object_to_string(object.clone()),
        DvmValue::USize(u) => u.to_string(),
        DvmValue::Array(a) => a.borrow().to_string(),
        DvmValue::ConstantPointer(p) => p.to_string(),
        DvmValue::Empty => String::from("Empty"),
    }
}

fn object_to_string(object: Rc<RefCell<DvmObject>>) -> String {
    if object.borrow().implementation().fqn() == "std::core::StringImpl" {
        extract_string_from_string(object.clone())
    } else {
        todo!("what happens if we don't have a String?")
    }
}

fn extract_string_from_string(string_object: Rc<RefCell<DvmObject>>) -> String {
    let object_ref = string_object.borrow();
    let bytes_field_value = object_ref.read_field("bytes");
    if let DvmValue::ConstantPointer(bytes_ptr) = bytes_field_value {
        let mut v = Vec::new();
        for i in 0..bytes_ptr.length {
            v.push(unsafe { bytes_ptr.raw_pointer.add(i).read() });
        }
        String::from_utf8(v).unwrap()
    } else {
        panic!("StringImpl.bytes field is not a DvmValue::ConstantPointer.");
    }
}
